import os
from cheap_repr import repr_str, cheap_repr
from birdseye import eye

path = os.path.join(os.path.expanduser('~'), '.birdseye_test.db')
os.environ.setdefault('BIRDSEYE_DB', 'sqlite:///' + path)

repr_str.maxparts = 30
cheap_repr.raise_exceptions = True

eye.num_samples['big']['list'] = 10



# coding=utf8

"""
Module docstrings before __future__ imports can break things...
"""

from __future__ import division

import ast
import json
import os
import random
import re
import sys
import unittest
import weakref
from collections import namedtuple
from copy import copy
from functools import partial
from importlib import import_module
from multiprocessing.dummy import Pool as ThreadPool
from time import sleep

from bs4 import BeautifulSoup
from cheap_repr import register_repr
from littleutils import file_to_json, only

from birdseye import eye
from birdseye.bird import NodeValue, is_interesting_expression, is_obvious_builtin
from birdseye.utils import PYPY
from tests.utils import SharedCounter
from collections.abc import Set, Mapping

Session = eye.db.Session
Call = eye.db.Call


@eye
def bar():
    pass


# noinspection PyStatementEffect
@eye()
def foo():
    x = 1
    y = 2
    if x + y > 5:
        1 / 0
    else:
        x * y
    try:
        bar(x + x, 2 / 0, y + y)
        foo
    except ZeroDivisionError:
        x - y
    for i in [1, 2]:
        for j in [3, 4]:
            i + j
        for k in [5]:
            k
    z = 0
    while z < 2:
        z += 1
        z ** z
    bar()
    {'list': [n for n in [1, 2]]}

    try:
        error()
    except ValueError:
        pass

    [1, 2, 3][:2]


@eye
def error():
    raise ValueError()


class NormalClass(object):
    def __init__(self):
        self.x = 1

    def __repr__(self):
        return '<A>'


class SlotClass(object):
    __slots__ = ('slot1',)

    def __init__(self):
        self.slot1 = 3

    def __repr__(self):
        return '<B>'


call_id = SharedCounter()


def call_id_mock(*_):
    return 'test_id_%s' % call_id.increment()


eye._call_id = call_id_mock


def get_call_ids(func):
    start_id = call_id.value + 1
    func()
    end_id = call_id.value + 1
    return ['test_id_%s' % i for i in range(start_id, end_id)]


def hydrate(call):
    str(call.function.name)
    return copy(call)


# Do this here to make call ids consistent
golden_calls = {
    name: [hydrate(Session().query(Call).filter_by(id=c_id).one())
           for c_id in get_call_ids(lambda: import_module('test_scripts.' + name))]
    for name in ('gold', 'traced')
}

CallStuff = namedtuple('CallStuff', 'call, soup, call_data, func_data')


@eye.db.provide_session
def get_call_stuff(sess, c_id):
    call = sess.query(Call).filter_by(id=c_id).one()

    # <pre> makes it preserve whitespace
    soup = BeautifulSoup('<pre>' + call.function.html_body + '</pre>', 'html.parser')

    call_data = normalise_call_data(call.data)
    func_data = json.loads(call.function.data)
    return CallStuff(copy(call), soup, call_data, func_data)


def byteify(x):
    # Legacy from Python 2 days
    return x


def normalise_call_data(call_data):
    """
    Replace type indices with type names.
    Sort type_names.
    :type call_data: str
    :rtype: dict
    """
    data = byteify(json.loads(call_data))
    types = data['type_names']

    def fix(x):
        if isinstance(x, dict):
            return dict((key, fix(value)) for key, value in x.items())
        elif isinstance(x, list):
            result = [x[0]]
            type_index = x[1]
            if type_index < 0:
                assert type_index in (-1, -2)
                result.append(type_index)
            else:
                result.append(types[type_index])
            result.append(x[2])

            for y in x[3:]:
                y[1] = fix(y[1])
                result.append(y)
            return result
        else:
            return x

    data['node_values'] = fix(data['node_values'], )
    data['type_names'].sort()
    return data


class TestBirdsEye(unittest.TestCase):
    maxDiff = None

    def test_stuff(self):
        call_ids = get_call_ids(foo)
        call, soup, call_data, func_data = get_call_stuff(call_ids[0])

        node_values = call_data['node_values']
        actual_values = {'expr': {}, 'stmt': {}, 'loop': {}}
        loops = {}
        actual_node_loops = {}
        for span in soup('span'):
            index = span['data-index']
            if index not in node_values:
                continue

            if 'loop' in span['class']:
                data_type = 'loop'
            elif 'stmt' in span['class']:
                data_type = 'stmt'
            else:
                data_type = 'expr'

            text = span.text.strip()
            actual_values[data_type][text] = node_values[index]
            if data_type == 'loop':
                loops[text.split()[1]] = index
            this_node_loops = func_data['node_loops'].get(index)
            if this_node_loops:
                actual_node_loops[text] = [str(x) for x in this_node_loops]

        def func_value(f):
            result = [repr(f), 'function', {}, ['__wrapped__', [repr(f.__wrapped__), 'function', {}]]]
            return result

        s = ['', -2, {}]

        expected_values = {
            'expr': {
                'x': ['1', 'int', {}],
                'y': ['2', 'int', {}],
                'x + y': ['3', 'int', {}],
                'x + y > 5': ['False', 'bool', {}],
                'x * y': ['2', 'int', {}],
                '2 / 0': ['ZeroDivisionError: division by zero', -1, {}],
                'bar': func_value(bar),
                'error': func_value(error),
                'bar()': ['None', 'NoneType', {'inner_calls': [call_ids[1]]}],
                'x + x': ['2', 'int', {}],
                'x - y': ['-1', 'int', {}],
                'i': {'0': {'0': ['1', 'int', {}],
                            '1': ['1', 'int', {}]},
                      '1': {'0': ['2', 'int', {}],
                            '1': ['2', 'int', {}]}},
                'i + j': {'0': {'0': ['4', 'int', {}],
                                '1': ['5', 'int', {}]},
                          '1': {'0': ['5', 'int', {}],
                                '1': ['6', 'int', {}]}},
                'j': {'0': {'0': ['3', 'int', {}],
                            '1': ['4', 'int', {}]},
                      '1': {'0': ['3', 'int', {}],
                            '1': ['4', 'int', {}]}},
                'k': {'0': {'0': ['5', 'int', {}]},
                      '1': {'0': ['5', 'int', {}]}},

                '[1, 2, 3][:2]': ['[1, 2]',
                                  'list',
                                  {'len': 2},
                                  ['0', ['1', 'int', {}]],
                                  ['1', ['2', 'int', {}]]],

                # These are the values of z as in z ** z, not z < 2
                'z': {'0': ['1', 'int', {}],
                      '1': ['2', 'int', {}]},

                'z ** z': {'0': ['1', 'int', {}],
                           '1': ['4', 'int', {}]},
                'z < 2': {'0': ['True', 'bool', {}],
                          '1': ['True', 'bool', {}],
                          '2': ['False', 'bool', {}]},
                '[n for n in [1, 2]]': ['[1, 2]', 'list',
                                        {'len': 2},
                                        ['0', ['1', 'int', {}]],
                                        ['1', ['2', 'int', {}]]],
                'n': {'0': ['1', 'int', {}],
                      '1': ['2', 'int', {}]},
                "{'list': [n for n in [1, 2]]}":
                    ["{'list': [1, 2]}", 'dict',
                     {'len': 1},
                     ["'list'",
                      ['[1, 2]', 'list',
                       {'len': 2},
                       ['0', ['1', 'int', {}]],
                       ['1', ['2', 'int', {}]]]]],
                'error()': ['ValueError', -1, {'inner_calls': [call_ids[2]]}],
            },
            'stmt': {
                'x = 1': s,
                'y = 2': s,
                '[1, 2, 3][:2]': s,
                '''
    if x + y > 5:
        1 / 0
    else:
        x * y
                '''.strip(): s,
                'x * y': s,
                '''
    try:
        bar(x + x, 2 / 0, y + y)
        foo
    except ZeroDivisionError:
        x - y
                '''.strip(): s,
                'bar(x + x, 2 / 0, y + y)': s,
                'x - y': s,
                'i + j': {'0': {'0': s, '1': s}, '1': {'0': s, '1': s}},
                'k': {'0': {'0': s}, '1': {'0': s}},
                'bar()': s,
                "{'list': [n for n in [1, 2]]}": s,
                'error()': s,
                '''
    try:
        error()
    except ValueError:
        pass
                '''.strip(): s,
                'pass': s,
                'z ** z': {'0': s, '1': s},
                'z += 1': {'0': s, '1': s},
                'z = 0': s,
            },
            'loop': {
                '''
    for i in [1, 2]:
        for j in [3, 4]:
            i + j
        for k in [5]:
            k
                '''.strip(): s,
                '''
        for j in [3, 4]:
            i + j
                '''.strip(): {'0': s, '1': s},
                '''
        for k in [5]:
            k
                '''.strip(): {'0': s, '1': s},
                'for n in [1, 2]': s,
                '''
    while z < 2:
        z += 1
        z ** z
                '''.strip(): s,
            }
        }
        self.assertEqual(byteify(actual_values), expected_values)

        expected_node_loops = {
            'z': [loops['z']],
            'z ** z': [loops['z']],
            'z += 1': [loops['z']],
            'z < 2': [loops['z']],
            'i + j': [loops['i'], loops['j']],
            'i': [loops['i'], loops['j']],
            'j': [loops['i'], loops['j']],
            'k': [loops['i'], loops['k']],
            '''
        for j in [3, 4]:
            i + j
            '''.strip(): [loops['i']],
            '''
        for k in [5]:
            k
            '''.strip(): [loops['i']],
            'n': [loops['n']]
        }
        self.assertEqual(actual_node_loops, expected_node_loops)

    def test_comprehension_loops(self):
        # noinspection PyUnusedLocal
        @eye
        def f():
            # @formatter:off
            for i in ([([x for x in [] for y in []], [x for x in [] for y in []]) for x in [x for x in [] for y in []] for y in [x for x in [] for y in []]], [([x for x in [] for y in []], [x for x in [] for y in []]) for x in [x for x in [] for y in []] for y in [x for x in [] for y in []]]):
                pass
            # @formatter:on

        soup = get_call_stuff(get_call_ids(f)[0]).soup
        for line in str(soup).splitlines():
            self.assertTrue(line.count('for') in (0, 1))

    def test_expansion(self):
        @eye
        def f():
            x = {'t': [(7, 8, [9, 10]), NormalClass(), SlotClass(), "Hello World!" * 50]}
            len(x)

        stuff = get_call_stuff(get_call_ids(f)[0])
        value = [x for x in stuff.call_data['node_values'].values()
                 if isinstance(x, list) and
                 "'t': " in x[0]
                 ][0]
        self.assertEqual(
            value,
            ["{'t': [(7, 8, [...]), <A>, <B>, 'Hello World!H...d!Hello World!']}",
             'dict', {'len': 1},
             ["'t'", ["[(7, 8, [9, 10]), <A>, <B>, 'Hello World!H...d!Hello World!']",
                      'list', {'len': 4},
                      ['0', ['(7, 8, [9, 10])',
                             'tuple', {'len': 3},
                             ['0', ['7', 'int', {}]],
                             ['1', ['8', 'int', {}]],
                             ['2', ['[9, 10]', 'list', {'len': 2}]]]],
                      ['1', ['<A>', 'NormalClass', {},
                             ['x', ['1', 'int', {}]]]],
                      ['2', ['<B>', 'SlotClass', {},
                             ['slot1', ['3', 'int', {}]]]],
                      ['3', ["'Hello World!H...d!Hello World!'",
                             'str', {'len': 600}]]]]])

    def test_against_files(self):

        @register_repr(weakref.ref)
        def repr_weakref(*_):
            return '<weakref>'

        def normalise_addresses(string):
            return re.sub(r'at 0x\w+>', 'at 0xABC>', string)

        for name, calls in golden_calls.items():
            data = [dict(
                arguments=byteify(json.loads(normalise_addresses(call.arguments))),
                return_value=byteify(normalise_addresses(str(call.return_value))),
                exception=call.exception,
                traceback=call.traceback,
                data=normalise_call_data(normalise_addresses(call.data)),
                function=dict(
                    name=byteify(call.function.name),
                    html_body=byteify(call.function.html_body),
                    lineno=call.function.lineno,
                    data=byteify(json.loads(call.function.data)),
                ),
            ) for call in calls]
            version = ('pypy' if PYPY else '') + '.'.join(map(str, sys.version_info[:2]))
            path = os.path.join(os.path.dirname(__file__), 'golden-files', version, name + '.json')

            if os.getenv("FIX_TESTS"):
                with open(path, 'w') as f:
                    json.dump(data, f, indent=2, sort_keys=True)
            else:
                self.assertEqual(data, byteify(file_to_json(path)))

    def test_decorate_class(self):
        with self.assertRaises(TypeError) as e:
            # noinspection PyUnusedLocal
            @eye
            class Testclass(object):
                def barz(self):
                    return 'class decorator test'

        self.assertEqual(str(e.exception),
                         'Decorating classes is no longer supported')

    def test_expand_exceptions(self):
        expand = partial(NodeValue.expression, eye.num_samples)

        class A(object):
            def __len__(self):
                assert 0

        with self.assertRaises(AssertionError):
            len(A())

        self.assertIsNone(expand(A(), 1).meta)
        self.assertEqual(expand([4, 4, 4], 1).meta['len'], 3)

        class FakeSet(Set):
            def __len__(self):
                return 0

            def __iter__(self):
                pass

            def __contains__(self, x):
                pass

        class B(FakeSet):
            def __iter__(self):
                assert 0

        with self.assertRaises(AssertionError):
            list(B())

        self.assertIsNone(expand(B(), 1).children)

        class C(FakeSet):
            def __iter__(self):
                yield 1
                yield 2
                assert 0

        def children_keys(cls):
            return [k for k, _ in expand(cls(), 1).children]

        with self.assertRaises(AssertionError):
            list(C())

        self.assertEqual(children_keys(C), ['<0>', '<1>'])

        class D(object):
            def __init__(self):
                self.x = 3
                self.y = 4

            def __getattribute__(self, item):
                assert item not in ['x', 'y']
                return object.__getattribute__(self, item)

        with self.assertRaises(AssertionError):
            str(D().x)

        # expand goes through __dict__ so x and y are reachable
        self.assertEqual(sorted(children_keys(D)), ['x', 'y'])

        class E(Mapping):
            def __len__(self):
                return 0

            def __getitem__(self, key):
                assert 0

            def __iter__(self):
                yield 4

        with self.assertRaises(AssertionError):
            list(E().items())

        self.assertIsNone(expand(E(), 1).children)

    def test_is_interesting_expression(self):
        def check(s):
            return is_interesting_expression(ast.parse(s, mode='eval').body)

        self.assertFalse(check('1'))
        self.assertFalse(check('-1'))
        self.assertTrue(check('-1-3'))
        self.assertFalse(check('"abc"'))
        self.assertTrue(check('abc'))
        self.assertFalse(check('[]'))
        self.assertFalse(check('[1, 2]'))
        self.assertFalse(check('[1, 2, "abc"]'))
        self.assertFalse(check('[[[]]]'))
        self.assertFalse(check('{}'))
        self.assertFalse(check('{1:2}'))
        self.assertFalse(check('["abc", 1, [2, {7:3}, {}, {3:[5, ["lkj"]]}]]'))
        self.assertTrue(check('["abc", 1+3, [2, {7:3}, {}, {3:[5, ["lkj"]]}]]'))

    def test_is_obvious_builtin(self):
        def check(s, value):
            return is_obvious_builtin(ast.parse(s, mode='eval').body, value)

        self.assertTrue(check('len', len))
        self.assertTrue(check('max', max))
        self.assertTrue(check('True', True))
        self.assertTrue(check('False', False))
        self.assertTrue(check('None', None))
        self.assertFalse(check('len', max))
        self.assertFalse(check('max', len))
        self.assertFalse(check('0', False))
        self.assertFalse(check('not True', False))

    def test_tracing_magic_methods(self):
        class A(object):
            @eye
            def __repr__(self):
                return '%s(label=%r)' % (self.__class__.__name__, self.label)

            @eye
            def __getattr__(self, item):
                return item

            @eye
            def __getattribute__(self, item):
                return object.__getattribute__(self, item)

            @eye
            def __len__(self):
                return self.length

        a = A()
        a.label = 'hello'
        a.length = 3

        @eye
        def test_A():
            self.assertEqual(a.label, 'hello')
            self.assertEqual(a.length, 3)
            self.assertEqual(a.thing, 'thing')
            self.assertEqual(repr(a), "A(label='hello')")

        test_A()

    def test_unicode(self):
        @eye
        def f():
            return u'é'

        self.assertEqual(f(), u'é')

    def test_optional_eye(self):
        @eye(optional=True)
        def f(x):
            return x * 3

        call_stuff = get_call_stuff(get_call_ids(lambda: f(2, trace_call=True))[0])
        self.assertEqual(call_stuff.call.result, '6')

        call = eye.enter_call
        eye.enter_call = lambda *args, **kwargs: 1 / 0
        try:
            self.assertEqual(f(3, trace_call=False), 9)
            self.assertEqual(f(4), 12)
        finally:
            eye.enter_call = call

    def test_first_check(self):
        def deco(f):
            f.attr = 3
            return f

        # Correct order, everything fine
        @deco
        @eye
        def baz():
            pass

        with self.assertRaises(ValueError):
            # @eye should notice it was applied second
            # because of __wrapped__ attribute
            @eye
            @deco
            def baz():
                pass

    def test_concurrency(self):
        ids = get_call_ids(lambda: ThreadPool(5).map(sleepy, range(25)))
        results = [int(get_call_stuff(i).call.result) for i in ids]
        self.assertEqual(sorted(results), list(range(0, 50, 2)))

    def test_middle_iterations(self):
        @eye
        def f():
            for i in range(20):
                for j in range(20):
                    if i == 10 and j >= 12:
                        str(i + 1)

        stuff = get_call_stuff(get_call_ids(f)[0])

        iteration_list = only(stuff.call_data['loop_iterations'].values())
        indexes = [i['index'] for i in iteration_list]
        self.assertEqual(indexes, [0, 1, 2, 10, 17, 18, 19])

        iteration_list = only(iteration_list[3]['loops'].values())
        indexes = [i['index'] for i in iteration_list]
        self.assertEqual(indexes, [0, 1, 2, 12, 13, 17, 18, 19])

    @classmethod
    def tearDownClass(cls):
        assert not eye.stack, eye.stack
        assert not eye.main_to_secondary_frames, eye.main_to_secondary_frames
        assert not eye.secondary_to_main_frames, eye.secondary_to_main_frames


@eye
def sleepy(x):
    sleep(random.random())
    return x * 2


if __name__ == '__main__':
    unittest.main()



import sys
from importlib import import_module

try:
    from .version import __version__
except ImportError:  # pragma: no cover
    # version.py is auto-generated with the git tag when building
    __version__ = "???"


# birdseye has so many dependencies that simply importing them can be quite slow
# Sometimes you just want to leave an import sitting around without actually using it
# These proxies ensure that if you do, program startup won't be slowed down
# In a nutshell:
#     from birdseye import eye
# is a lazy version of
#     from birdseye.bird import eye

class _SimpleProxy(object):
    def __init__(self, val):
        object.__setattr__(self, '_SimpleProxy__val', val)

    def __call__(self, *args, **kwargs):
        return self.__val()(*args, **kwargs)

    def __getattr__(self, item):
        return getattr(self.__val(), item)

    def __setattr__(self, key, value):
        setattr(self.__val(), key, value)


eye = _SimpleProxy(lambda: import_module('birdseye.bird').eye)
BirdsEye = _SimpleProxy(lambda: import_module('birdseye.bird').BirdsEye)


def load_ipython_extension(ipython_shell):
    from birdseye.ipython import BirdsEyeMagics
    ipython_shell.register_magics(BirdsEyeMagics)


if sys.version_info.major == 3:
    from birdseye.import_hook import BirdsEyeFinder

    sys.meta_path.insert(0, BirdsEyeFinder())



import sys
import unittest

from tests.utils import requires_python_version


class TestTreeTrace(unittest.TestCase):
    maxDiff = None

    @requires_python_version(3.5)
    def test_async_forbidden(self):
        from birdseye.tracer import TreeTracerBase
        tracer = TreeTracerBase()
        with self.assertRaises(ValueError):
            exec("""
@tracer
async def f(): pass""")

        if sys.version_info >= (3, 6):
            with self.assertRaises(ValueError):
                exec("""
@tracer
async def f(): yield 1""")



import birdseye.trace_module_deep


def deco(f):
    return f


def m():
    qwe = 9
    str(qwe)

    class A:
        for i in range(3):
            str(i * i)

        class B:
            str([[i * 2 for i in range(j)]
                 for j in range(3)])

        (lambda *_: 9)(None)
        (lambda x: [i * x for i in range(3)])(8)
        str({(lambda x: i + x)(7) for i in range(3)})

        @deco
        def foo(self):
            x = 9 * 0
            str(1 + 2 + x)
            return self

        def bar(self):
            return 1 + 3

    A().foo().bar()


m()



import unittest
from time import sleep

from littleutils import only
from selenium import webdriver
from selenium.webdriver import ActionChains
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By

from birdseye import eye


@eye
def foo():
    for i in range(20):
        for j in range(3):
            int(i * 13 + j * 17)
            if i > 0:
                try:
                    assert j
                except AssertionError:
                    pass
    str(bar())

    x = list(range(1, 30, 2))
    list(x)


@eye
def bar():
    pass


class TestInterface(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        chrome_options = Options()
        chrome_options.add_argument("--headless")
        chrome_options.add_argument("--no-sandbox")
        self.driver = webdriver.Chrome(options=chrome_options)
        self.driver.set_window_size(1600, 1200)
        self.driver.implicitly_wait(2)

    def test(self):
        try:
            self._do_test()
        except:
            self.driver.save_screenshot('error_screenshot.png')
            raise

    def _do_test(self):
        foo()
        driver = self.driver

        # On the index page, note the links to the function and call
        driver.get('http://localhost:7777/')
        function_link = driver.find_element(By.LINK_TEXT, 'foo')
        function_url = function_link.get_attribute('href')
        call_url = function_link.find_element(By.XPATH, '..//i/..').get_attribute('href')

        # On the file page, check that the links still match
        driver.find_element(By.PARTIAL_LINK_TEXT, 'test_interface').click()
        function_link = driver.find_element(By.LINK_TEXT, 'foo')
        self.assertEqual(function_link.get_attribute('href'), function_url)
        self.assertEqual(call_url, function_link.find_element(By.XPATH, '..//i/..').get_attribute('href'))

        # Finally navigate to the call and check the original call_url
        function_link.click()
        driver.find_element(By.CSS_SELECTOR, 'table a').click()
        self.assertEqual(driver.current_url, call_url)

        # Test hovering, clicking on expressions, and stepping through loops

        vals = {'i': 0, 'j': 0}
        exprs = driver.find_elements(By.CLASS_NAME, 'has_value')
        expr_value = driver.find_element(By.ID, 'box_value')

        expr_strings = [
            'i * 13 + j * 17',
            'j * 17',
            'i * 13',
        ]

        def find_by_text(text, elements):
            return only(n for n in elements if n.text == text)

        def find_expr(text):
            return find_by_text(text, exprs)

        def tree_nodes(root=driver):
            return root.find_elements(By.CLASS_NAME, 'jstree-node')

        def select(node, prefix, value_text):
            self.assertIn('box', classes(node))
            self.assertIn('has_value', classes(node))
            self.assertNotIn('selected', classes(node))
            node.click()
            self.assertIn('selected', classes(node))
            self.assertEqual(expr_value.text, value_text)
            tree_node = tree_nodes()[-1]
            self.assertEqual(tree_node.text, prefix + value_text)
            return tree_node

        def classes(node):
            return set(node.get_attribute('class').split())

        def assert_classes(node, *cls):
            self.assertEqual(classes(node), set(cls))

        for i, expr in enumerate(expr_strings):
            find_expr(expr).click()

        def step(loop, increment):
            selector = '.loop-navigator > .btn:%s-child' % ('first' if increment == -1 else 'last')
            buttons = driver.find_elements(By.CSS_SELECTOR, selector)
            self.assertEqual(len(buttons), 2)
            buttons[loop].click()
            vals['ij'[loop]] += increment

            for expr in expr_strings:
                ActionChains(driver).move_to_element(find_expr(expr)).perform()
                value = str(eval(expr, {}, vals))
                self.assertEqual(expr_value.text, value)
                node = only(n for n in tree_nodes()
                            if n.text.startswith(expr + ' ='))
                self.assertEqual(node.text, '%s = int: %s' % (expr, value))

        stmt = find_by_text('assert j', driver.find_elements(By.CLASS_NAME, 'stmt'))
        assert_classes(stmt, 'stmt', 'stmt_uncovered', 'box')

        step(0, 1)
        select(stmt, 'assert j : ', 'AssertionError')
        assert_classes(stmt, 'stmt', 'selected', 'box', 'hovering', 'has_value', 'exception_node')
        step(1, 1)
        self.assertEqual(tree_nodes()[-1].text, 'assert j : fine')
        assert_classes(stmt, 'stmt', 'selected', 'box', 'has_value', 'value_none')
        step(1, 1)
        step(0, -1)
        self.assertTrue({'stmt', 'stmt_uncovered', 'selected', 'box'} <= classes(stmt))
        step(1, -1)

        # Expanding values
        x_node = find_expr('x')
        tree_node = select(x_node, 'x = list: ', '[1, 3, 5, ..., 25, 27, 29]')
        tree_node.find_element(By.CLASS_NAME, 'jstree-ocl').click()  # expand
        sleep(0.2)
        self.assertEqual([n.text for n in tree_nodes(tree_node)],
                         ['len() = 15',
                          '0 = int: 1',
                          '1 = int: 3',
                          '2 = int: 5',
                          '3 = int: 7',
                          '4 = int: 9',
                          '10 = int: 21',
                          '11 = int: 23',
                          '12 = int: 25',
                          '13 = int: 27',
                          '14 = int: 29']),

        # Click on an inner call
        find_expr('bar()').find_element(By.CLASS_NAME, 'inner-call').click()
        self.assertEqual(driver.find_element(By.TAG_NAME, 'h2').text,
                         'Call to function: bar')



from __future__ import annotations

import sys
from importlib.machinery import ModuleSpec
from importlib.util import spec_from_loader
import ast
from types import ModuleType
from typing import Sequence, Iterator, cast


class BirdsEyeLoader:
    def __init__(self, spec, source, deep):
        self._spec = spec
        self.source = source
        self.deep = deep

    def create_module(self, spec):
        pass

    def exec_module(self, module):
        from birdseye.bird import eye
        eye.exec_string(
            source=self.source,
            filename=self._spec.origin,
            globs=module.__dict__,
            locs=module.__dict__,
            deep=self.deep,
        )

    def get_filename(self, fullname):
        return self._spec.loader.get_filename(fullname)

    def is_package(self, fullname):
        return self._spec.loader.is_package(fullname)


class BirdsEyeFinder:
    """Loads a module and looks for tracing inside, only providing a loader
    if it finds some.
    """

    def _find_plain_specs(
        self, fullname: str, path: Sequence[str] | None, target: ModuleType | None
    ) -> Iterator[ModuleSpec]:
        """Yield module specs returned by other finders on `sys.meta_path`."""
        for finder in sys.meta_path:
            # Skip this finder or any like it to avoid infinite recursion.
            if isinstance(finder, BirdsEyeFinder):
                continue

            try:
                plain_spec = finder.find_spec(fullname, path, target)
            except Exception:  # pragma: no cover
                continue

            if plain_spec:
                yield plain_spec

    def find_spec(
        self, fullname: str, path: Sequence[str] | None, target: ModuleType | None = None
    ) -> ModuleSpec | None:
        """This is the method that is called by the import system.

        It uses the other existing meta path finders to do most of the standard work,
        particularly finding the module's source code.
        If it finds a module spec, it returns a new spec that uses the BirdsEyeLoader.
        """
        for plain_spec in self._find_plain_specs(fullname, path, target):
            # Not all loaders have get_source, but it's an abstract method of the standard ABC InspectLoader.
            # In particular it's implemented by `importlib.machinery.SourceFileLoader`
            # which is provided by default.
            get_source = getattr(plain_spec.loader, 'get_source', None)
            if not callable(get_source):  # pragma: no cover
                continue

            try:
                source = cast(str, get_source(fullname))
            except Exception:  # pragma: no cover
                continue

            if not source:
                continue

            if "birdseye" not in source:
                return None

            deep, trace_stmt = should_trace(source)

            if not trace_stmt:
                return None

            loader = BirdsEyeLoader(plain_spec, source, deep)
            return spec_from_loader(fullname, loader)


def should_trace(source):
    trace_stmt = None
    deep = False
    for stmt in ast.parse(source).body:
        if isinstance(stmt, ast.Import):
            for alias in stmt.names:
                if alias.name.startswith('birdseye.trace_module'):
                    trace_stmt = stmt
                    if alias.name.endswith('deep'):
                        deep = True

        if isinstance(stmt, ast.ImportFrom) and stmt.module == 'birdseye':
            for alias in stmt.names:
                if alias.name.startswith('trace_module'):
                    trace_stmt = stmt
                    if alias.name.endswith('deep'):
                        deep = True
    return deep, trace_stmt



import ast
import json
import linecache
import ntpath
import os
import sys
import token

from littleutils import strip_required_prefix

from typing import Union, List, Any, Iterator, Tuple, Iterable


PYPY = 'pypy' in sys.version.lower()
IPYTHON_FILE_PATH = 'IPython notebook or shell'
FILE_SENTINEL_NAME = '$$__FILE__$$'


def path_leaf(path):
    # type: (str) -> str
    # http://stackoverflow.com/a/8384788/2482744
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)


def common_ancestor(paths):
    # type: (List[str]) -> str
    """
    Returns a path to a directory that contains all the given absolute paths
    """
    prefix = os.path.commonprefix(paths)

    # Ensure that the prefix doesn't end in part of the name of a file/directory
    prefix = ntpath.split(prefix)[0]

    # Ensure that it ends with a slash
    first_char_after = paths[0][len(prefix)]
    if first_char_after in r'\/':
        prefix += first_char_after

    return prefix


def short_path(path, all_paths):
    # type: (str, List[str]) -> str
    if path == IPYTHON_FILE_PATH:
        return path

    all_paths = [f for f in all_paths
                 if f != IPYTHON_FILE_PATH]
    prefix = common_ancestor(all_paths)
    if prefix in r'\/':
        prefix = ''
    return strip_required_prefix(path, prefix) or path_leaf(path)


def fix_abs_path(path):
    if path == IPYTHON_FILE_PATH:
        return path
    if os.path.sep == '/' and not path.startswith('/'):
        path = '/' + path
    return path


def of_type(type_or_tuple, iterable):
    # type: (Union[type, Tuple[Union[type, tuple], ...]], Iterable[Any]) -> Iterator[Any]
    return (x for x in iterable if isinstance(x, type_or_tuple))


def one_or_none(expression):
    """Performs a one_or_none on a sqlalchemy expression."""
    if hasattr(expression, 'one_or_none'):
        return expression.one_or_none()
    result = expression.all()
    if len(result) == 0:
        return None
    elif len(result) == 1:
        return result[0]
    else:
        raise Exception("There is more than one item returned for the supplied filter")


def flatten_list(lst):
    result = []
    for x in lst:
        if isinstance(x, list):
            result.extend(flatten_list(x))
        else:
            result.append(x)
    return result


def is_lambda(f):
    try:
        code = f.__code__
    except AttributeError:
        return False
    return code.co_name == (lambda: 0).__code__.co_name


class ProtocolEncoder(json.JSONEncoder):
    def default(self, o):
        try:
            method = o.as_json
        except AttributeError:
            return super(ProtocolEncoder, self).default(o)
        else:
            return method()


def read_source_file(filename):
    lines = linecache.getlines(filename)
    return ''.join(lines)


def source_without_decorators(tokens, function_node):
    def_token = tokens.find_token(function_node.first_token, token.NAME, 'def')
    startpos = def_token.startpos
    source = tokens.text[startpos:function_node.last_token.endpos].rstrip()
    assert source.startswith('def')

    return startpos, source


def prn(*args):
    for arg in args:
        print(arg)
    if len(args) == 1:
        return args[0]
    return args


def is_ipython_cell(filename):
    return filename.startswith('<ipython-input-')


def is_future_import(node):
    return isinstance(node, ast.ImportFrom) and node.module == "__future__"


def get_unfrozen_datetime():
    try:
        # if freezegun could be active, we need to use real_datetime to ensure we use the actual time instead of the
        # time set by freezegun.
        # we have to import this at the last possible moment because birdeye is very likely to be imported before
        # freezegun is activated.
        from freezegun.api import real_datetime
    except ImportError:
        from datetime import datetime as real_datetime

    return real_datetime.now()


html_escape_table = {
    "&": "&amp;",
    '"': "&quot;",
    "'": "&#x27;",
    ">": "&gt;",
    "<": "&lt;",
}


def html_escape(text):
    return "".join(html_escape_table.get(c, c) for c in text)


def format_pandas_index(index):
    """
    Supports different versions of pandas
    """
    try:
        return index.astype(str)
    except Exception:
        try:
            return index.format(sparsify=False)
        except TypeError:
            return index.format()



from itertools import islice

from birdseye import eye

G = 9


@eye
def factorial(n):
    if n <= 1:
        return 1
    return n * factorial(n - 1)


def dummy(*args):
    pass


class SlotClass(object):
    __slots__ = ('slot1',)

    def __init__(self):
        self.slot1 = 3


@eye
def complex_args(pos1, pos2, key1=3, key2=4, *args, **kwargs):
    return [pos1, pos2, kwargs]


@eye
def gen():
    for i in range(6):
        yield i


@eye
def use_gen_1(g):
    for x in islice(g, 3):
        dummy(x)


@eye
def use_gen_2(g):
    for y in g:
        dummy(y)


class MyClass(object):
    @eye
    def __add__(self, other):
        return other

    @eye
    def __enter__(self):
        pass

    @eye
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


@eye
def main():
    assert factorial(3) == 6

    vals = []
    for i in range(100):
        vals.append([])
        for j in range(2 * i):
            vals[-1].append(i + j)
            dummy(vals)

    for i in range(6):
        try:
            dummy(1 / (i % 2) + 10)
        except ZeroDivisionError:
            continue
        if i == 3:
            break

    c = MyClass() + MyClass()
    c.list = [[x + y for x in range(100)] for y in range(100)]
    sum  (n for n in range(4))
    dummy({n for n in range(4)})
    dummy({n: n for n in range(1)})
    with c:
        pass
    dummy(c + SlotClass())

    assert complex_args(
        list(range(1000)),
        "hello",
        key2=8,
        kwarg1={'key': 'value'}
    ) == [list(range(1000)),
          'hello',
          dict(kwarg1={'key': 'value'})]

    assert complex_args(*[1, 2], **{'k': 23}) == [1, 2, {'k': 23}]

    assert eval('%s + %s' % (1, 2)) == 3

    x = 1
    x += 5
    assert x == 6
    del x

    dummy(True, False, None)

    assert [1, 2, 3][1] == 2
    assert (1, 2, 3)[:2] == (1, 2)

    try:
        raise ValueError()
    except AssertionError as e:
        pass
    except TypeError:
        pass
    except:
        pass
    finally:
        dummy()

    while True:
        break

    assert (lambda x: x * 2)(4) == 8

    global G
    G = 4
    assert G == 4

    g = gen()
    use_gen_1(g)
    use_gen_2(g)


main()



# coding=utf8
import unittest

from tests.utils import requires_python_version


class TestImportHook(unittest.TestCase):
    @requires_python_version(3.5)
    def test_should_trace(self):
        from birdseye.import_hook import should_trace
        deep, trace_stmt = should_trace('import birdseye.trace_module')
        self.assertFalse(deep)
        self.assertIsNotNone(trace_stmt)

        deep, trace_stmt = should_trace('import birdseye.trace_module_deep')
        self.assertTrue(deep)
        self.assertIsNotNone(trace_stmt)

        deep, trace_stmt = should_trace('from birdseye import trace_module_deep, eye')
        self.assertTrue(deep)
        self.assertIsNotNone(trace_stmt)

        deep, trace_stmt = should_trace('from birdseye import trace_module, eye')
        self.assertFalse(deep)
        self.assertIsNotNone(trace_stmt)

        deep, trace_stmt = should_trace('from birdseye import eye')
        self.assertFalse(deep)
        self.assertIsNone(trace_stmt)


if __name__ == '__main__':
    unittest.main()



import unittest

from birdseye.db import Database


class TestDatabase(unittest.TestCase):
    def test_key_value_store(self):
        kv = Database().key_value_store

        self.assertIsNone(kv.thing)

        kv.thing = 'foo'
        self.assertEqual(kv.thing, 'foo')

        kv.thing = 'bar'
        self.assertEqual(kv.thing, 'bar')


if __name__ == '__main__':
    unittest.main()
